import typing as t
from http import HTTPMethod

from httpx import AsyncClient, HTTPStatusError, RequestError, Response
from loguru import logger

from ash_utils import constants
from ash_utils.middlewares import request_id_var


class BaseApi:
    class ThirdPartyRequestError(Exception):
        def __init__(self, message: str):
            self.message = message

    class ThirdPartyHttpStatusError(Exception):
        def __init__(self, status_code: int, text: str, response: Response) -> None:
            self.status_code = status_code
            self.text = text
            self.response = response

    def __init__(self, client: AsyncClient, request_id_header_name: str = constants.REQUEST_ID_HEADER_NAME):
        self.client = client
        self.request_id_header_name = request_id_header_name

    async def _send_request(
        self,
        method: HTTPMethod,
        url: str,
        body: dict[str, t.Any] | list[int] | None = None,
        data: t.Any | None = None,
        params: dict[str, t.Any] | None = None,
        headers: dict | None = None,
    ) -> Response:
        with logger.contextualize(url=url, method=method, params=params):
            if headers is None:
                headers = {self.request_id_header_name: request_id_var.get()}
            else:
                headers[self.request_id_header_name] = request_id_var.get()

            logger.info("Send request")

            try:
                response = await self.client.request(
                    method=method,
                    url=url,
                    json=body,
                    data=data,
                    params=params,
                    headers=headers,
                    follow_redirects=True,
                )
            except RequestError as ex:
                raise self.ThirdPartyRequestError(message=str(ex)) from ex

            try:
                response.raise_for_status()
            except HTTPStatusError as ex:
                raise self.ThirdPartyHttpStatusError(
                    status_code=response.status_code,
                    text=response.text,
                    response=response,
                ) from ex

            logger.info(f"Response received | {response.status_code=}")

            return response



from ash_utils.middlewares.catch_unexpected_exception import CatchUnexpectedExceptionsMiddleware
from ash_utils.middlewares.permissions_policy import PermissionsPolicy
from ash_utils.middlewares.request_id import RequestIDMiddleware, request_id_var
from ash_utils.middlewares.security import configure_security_headers

__all__ = [
    "CatchUnexpectedExceptionsMiddleware",
    "PermissionsPolicy",
    "RequestIDMiddleware",
    "configure_security_headers",
    "request_id_var",
]



import json
import re
from typing import cast

from fastapi import Request, status
from fastapi.responses import JSONResponse
from loguru import logger
from starlette.types import ASGIApp, Receive


class CatchUnexpectedExceptionsMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        response_error_message: str,
        response_status_code: int = status.HTTP_500_INTERNAL_SERVER_ERROR,
        context_keys: list[str] | None = None,
    ):
        self.app = app
        self.response_error_message = response_error_message
        self.response_status_code = response_status_code
        self.context_keys = context_keys or []

    async def __call__(self, scope, receive, send):
        if scope["type"] != "http":  # pragma: no cover
            await self.app(scope, receive, send)
            return

        receive_proxy = ReceiveProxy(receive=receive)

        try:
            await self.app(scope, cast(Receive, receive_proxy), send)
        except Exception:
            request = Request(scope, cast(Receive, receive_proxy))
            context = await self._extract_request_info(request, receive_proxy)
            with logger.contextualize(**context):
                logger.exception(f"Unexpected exception. Url: {request.url}")
            response = JSONResponse(
                status_code=self.response_status_code,
                content={"detail": self.response_error_message},
            )
            await response(scope, receive, send)
            return

    async def _extract_request_info(self, request: Request, receive_proxy: "ReceiveProxy") -> dict[str, str]:
        """
        Extracts specified keys from the request data or query parameters.

        Args:
            request: The incoming request object.

        Returns:
            A dictionary containing the extracted key-value pairs.
        """
        try:
            body = await self._get_request_body(request, receive_proxy)
            data = json.loads(body)
        except Exception:
            data = cast(dict, request.query_params)

        context = {}
        for key in self.context_keys:
            value = _find_key_in_dict(data, key)
            if value is not None:
                context[_to_snake(key)] = value
        return context

    async def _get_request_body(self, request: Request, receive_proxy: "ReceiveProxy") -> bytes:
        """
        Returns the request body, either from the cache or by reading from the stream.
        This is necessary because the request body can only be read once.

        Args:
            request: The incoming request object.
            receive_proxy: The proxy object to handle the request body.
        """

        if not receive_proxy.has_body():
            logger.debug("Request body not cached, consuming it now.")
            return await request.body()

        logger.debug("Request body already cached, using cached value.")
        return receive_proxy.cached_body


def _find_key_in_dict(data: dict, key: str) -> str | None:
    """
    Finds the value of a specified key in a nested dictionary.

    Args:
        data: The dictionary to search.
        key: The key to find.

    Returns:
        The value associated with the key, or None if not found.
    """
    if key in data:
        return data[key]
    for value in data.values():
        if isinstance(value, dict):
            found = _find_key_in_dict(value, key)
            if found is not None:
                return found
    return None


def _to_snake(camel: str) -> str:
    """Convert a PascalCase, camelCase, or kebab-case string to snake_case.

    Args:
        camel: The string to convert.

    Returns:
        The converted string in snake_case.
    """
    # Handle the sequence of uppercase letters followed by a lowercase letter
    snake = re.sub(r"([A-Z]+)([A-Z][a-z])", lambda m: f"{m.group(1)}_{m.group(2)}", camel)
    # Insert an underscore between a lowercase letter and an uppercase letter
    snake = re.sub(r"([a-z])([A-Z])", lambda m: f"{m.group(1)}_{m.group(2)}", snake)
    # Insert an underscore between a digit and an uppercase letter
    snake = re.sub(r"([0-9])([A-Z])", lambda m: f"{m.group(1)}_{m.group(2)}", snake)
    # Insert an underscore between a lowercase letter and a digit
    snake = re.sub(r"([a-z])([0-9])", lambda m: f"{m.group(1)}_{m.group(2)}", snake)
    # Replace hyphens with underscores to handle kebab-case
    snake = snake.replace("-", "_")
    return snake.lower()


class ReceiveProxy:
    """Class that caches the request body so that it can be used later."""

    def __init__(self, receive: Receive):
        self._receive = receive
        self.cached_body = b""
        self._consumed = False

    async def __call__(self):
        message = await self._receive()
        if message["type"] == "http.request":
            chunk = message.get("body", b"")
            self.cached_body += chunk
            if not message.get("more_body", False):
                self._consumed = True
        return message

    def has_body(self) -> bool:
        """Check if the body has been consumed."""
        return self._consumed



import json
from functools import partial

import sentry_sdk
from loguru import logger
from nested_lookup import nested_update
from sentry_sdk.integrations.loguru import LoguruIntegration
from sentry_sdk.scrubber import DEFAULT_DENYLIST, DEFAULT_PII_DENYLIST, EventScrubber
from sentry_sdk.types import Event

from ash_utils.integrations.constants import KEYS_TO_FILTER, REDACTION_STRING, SENSITIVE_DATA_FLAG, LoguruConfigs


def _redact_logentry(event: Event) -> Event:
    """Redacts sensitive errors from the log entry before sending to Sentry."""

    if "logentry" in event:
        logentry_string = json.dumps(event["logentry"])
        extra = event.get("extra", {}).get("extra", {})

        if SENSITIVE_DATA_FLAG in logentry_string:
            event["logentry"]["message"] = f"REDACTED SENSITIVE ERROR | {extra.get('kit_id')}"  # type: ignore[reportIndexIssue]
        else:
            for key in KEYS_TO_FILTER:
                if key in logentry_string:
                    event["logentry"]["message"] = f"REDACTED SENSITIVE ERROR | key: {key} | {extra.get('kit_id')}"  # type: ignore[reportIndexIssue]
                    break
    return event


def _try_parse_json(data_string: str) -> dict | None:
    """Attempts to parse a string as JSON. Returns a dictionary if successful, otherwise None."""

    try:
        return json.loads(data_string.replace("'", '"'))
    except json.JSONDecodeError:
        return None


def _redact_exception(event: Event) -> Event:
    """Redacts sensitive-tagged values or values of `keys_to_filter` in exception details."""

    for values in event.get("exception", {}).get("values", []):
        exception_value = values.get("value")
        if not exception_value:
            continue

        if SENSITIVE_DATA_FLAG in exception_value:
            values["value"] = REDACTION_STRING
            continue

        try:
            exception_value_dict = _try_parse_json(exception_value)
            if exception_value_dict:
                for key in KEYS_TO_FILTER:
                    nested_update(
                        exception_value_dict,
                        key=key,
                        value=REDACTION_STRING,
                        in_place=True,
                    )
                values["value"] = json.dumps(exception_value_dict)
            elif any(key in exception_value for key in KEYS_TO_FILTER):
                values["value"] = REDACTION_STRING
        except Exception:
            logger.warning("Unhandled error encountered while redacting the exception for a Sentry issue.")
            return _remove_potential_exception_pii(event)
    return event


def _remove_potential_exception_pii(event: Event) -> Event:
    """Removes potential PII from the exception context in the Sentry event.
    Only runs if the `_redact_exception` function fails
    """
    if "exception" in event and isinstance(event["exception"], dict):
        error_type = event["exception"]["values"][0]["type"]
        event["exception"] = {"values": [{}]}
        event["exception"]["values"][0]["type"] = error_type
    for key in ["contexts", "extra", "breadcrumbs", "tags"]:
        event[key] = {}
    return event


def before_send(event: Event, _hint: dict) -> Event:
    """Processes an event before sending to Sentry by redacting sensitive information.

    Args:
        event (Event): The Sentry event to be scrubbed.
        _hint (dict): optional dictionary containing information about the event (unused).

    Returns:
        Event: The redacted Sentry event
    """
    event_log_redacted = _redact_logentry(event)
    return _redact_exception(event_log_redacted)


def initialize_sentry(
    sentry_dsn: str,
    environment: str,
    release: str,
    traces_sample_rate: float = 0.1,
    sample_rate: float = 1.0,
    additional_integrations: list | None = None,
    context_keys: list[str] | None = None,
):
    """Initializes the Sentry SDK with the provided configuration.

    #### Params:
        `sentry_dsn` (str): The DSN for the Sentry project.
        `environment` (str): The environment for the Sentry project.
        `release` (str): The release version for the Sentry project.
        `traces_sample_rate` (float): OPTIONAL - The sample rate for Sentry traces;
            defaults to 0.1 if not passed.
        `sample_rate` (float): OPTIONAL - The sample rate for Sentry events;
            defaults to 1.0 if not passed.
        `additional_integrations` (list): OPTIONAL - Additional Sentry integrations to include;
            integrations defaults to LoguruIntegration() if not passed.
        `context_keys` (list): OPTIONAL - List of keys to include in the error message;
            defaults to `["code", "kit_id", "event"]` if not passed.

    #### Defaults Applied Automatically:
    - `include_local_variables`: Set to `False` for security reasons.
    - `send_default_pii`: Disabled (`False`) to avoid sending user PII.
    - `Event Scrubber`: Uses an internal scrubber to filter sensitive data;
        custom denylist added to Sentry default PII denylist.
    - `before_send`: function to sanitize logs/exceptions in case EventScrubber misses anything.

    Example usage:
    ```python
    from ash_utils.integrations.sentry import initialize_sentry
    from sentry_sdk.integrations.fastapi import FastAPIIntegration
    from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration

    initialize_sentry(
        dsn="your-dsn",
        environment="staging",
        release="1.2.3",
        traces_sample_rate=0.5,
        additional_integrations=[
            FastAPIIntegration(), SqlalchemyIntegration()
        ],
    )
    ```
    """

    if not context_keys:
        context_keys = ["code", "kit_id", "event"]

    default_integrations = [
        LoguruIntegration(
            event_format=partial(LoguruConfigs.event_log_format, context_keys=context_keys),
            breadcrumb_format=LoguruConfigs.breadcrumb_log_format,
        ),
    ]
    if additional_integrations:
        default_integrations.extend(additional_integrations)

    # EventScrubber will merge pii_denylist with the default denylist at runtime
    custom_pii_denylist = KEYS_TO_FILTER + DEFAULT_PII_DENYLIST

    sentry_sdk.init(
        dsn=sentry_dsn,
        traces_sample_rate=traces_sample_rate,
        sample_rate=sample_rate,
        integrations=default_integrations,
        release=release,
        environment=environment,
        include_local_variables=False,
        send_default_pii=False,
        event_scrubber=EventScrubber(
            recursive=True,
            denylist=DEFAULT_DENYLIST,
            pii_denylist=custom_pii_denylist,
        ),
        before_send=before_send,
    )



from pydantic.alias_generators import to_snake

# Sentry Constants
REDACTION_STRING = "REDACTED"
SENSITIVE_DATA_FLAG = "SENSITIVE"
KEYS_TO_FILTER = [
    "address",
    "address1",
    "address2",
    "city",
    "country",
    "dob",
    "email",
    "first_name",
    "firstName",
    "last_name",
    "lastName",
    "password",
    "patient_address1",
    "patient_address2",
    "patient_city",
    "patient_email",
    "patient_state",
    "patient_zip",
    "patientAddress1",
    "patientAddress2",
    "patientCity",
    "patientEmail",
    "patientState",
    "patientZip",
    "PatientZip",
    "phone",
    "searchKeyword",
    "search_keyword",
    "shipping_address1",
    "shipping_address2",
    "shipping_city",
    "shipping_email",
    "shipping_state",
    "shipping_zip",
    "shippingAddress1",
    "shippingAddress2",
    "shippingCity",
    "shippingEmail",
    "shippingState",
    "shippingZip",
    "state",
    "zip",
]


class LoguruConfigs:
    """
    Holds Loguru format methods and error code constant
    """

    ASH_SYSTEM_ERROR_CODE = "ash-system-error"

    @staticmethod
    def breadcrumb_log_format(_):
        """
        Returns the log string loguru needs to format the LogRecord
        object to generate a log message.

        :param _: The record object (UNUSED).
        """
        return "{message} | {extra}"

    @staticmethod
    def event_log_format(record, context_keys: list[str]):
        """
        Returns a formatted string for loguru events.
        :param record: The record object.
        """

        if "code" not in context_keys:
            context_keys.insert(0, "code")

        if "code" not in record["extra"] and (exception_info := record.get("exception")):
            exception = exception_info[1]
            record["extra"]["code"] = getattr(exception, "code", LoguruConfigs.ASH_SYSTEM_ERROR_CODE)
        else:
            record["extra"]["code"] = record["extra"].get("code") or LoguruConfigs.ASH_SYSTEM_ERROR_CODE

        format_str = ""
        for key in context_keys:
            if value := record["extra"].get(to_snake(key)):
                format_str += f"[{value}] "

        format_str += "{message}"
        return format_str



from ash_utils.integrations.constants import KEYS_TO_FILTER
from ash_utils.integrations.sentry import before_send, initialize_sentry

__all__ = [
    "KEYS_TO_FILTER",
    "before_send",
    "initialize_sentry",
]



from starlette.datastructures import MutableHeaders
from starlette.types import ASGIApp, Receive, Scope, Send


class PermissionsPolicy:
    def __init__(self, app: ASGIApp, Option: dict[str, list[str]]):  # noqa: N803
        self.app = app
        self.allowed_policies = [
            "accelerometer",
            "ambient-light-sensor",
            "attribution-reporting",
            "autoplay",
            "bluetooth",
            "browsing-topics",
            "camera",
            "compute-pressure",
            "display-capture",
            "document-domain",
            "encrypted-media",
            "fullscreen",
            "geolocation",
            "gyroscope",
            "hid",
            "identity-credentials-get",
            "idle-detection",
            "local-fonts",
            "magnetometer",
            "microphone",
            "midi",
            "otp-credentials",
            "payment",
            "picture-in-picture",
            "publickey-credentials-create",
            "publickey-credentials-get",
            "screen-wake-lock",
            "serial",
            "storage-access",
            "usb",
            "web-share",
            "window-management",
            "xr-spatial-tracking",
        ]
        self.policy_header = self._generate_header_value(Option)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":  # pragma: no cover
            await self.app(scope, receive, send)
            return

        async def send_wrapper(message) -> None:
            if message["type"] == "http.response.start":
                headers = MutableHeaders(scope=message)
                headers.append("Permissions-Policy", self.policy_header)
            await send(message)

        await self.app(scope, receive, send_wrapper)

    def _generate_header_value(self, policy_dict: dict) -> str:
        policy_parts = []

        for feature, origins in policy_dict.items():
            if feature not in self.allowed_policies:
                raise ValueError(f"Invalid policy feature: {feature}")

            if not origins:
                policy_parts.append(f"{feature}=()")
                continue

            processed = []
            for origin in origins:
                origin_lower = origin.strip().lower()
                if origin_lower in {"self", "*", "src"}:
                    processed.append(f"'{origin_lower}'")
                else:
                    if not origin.startswith(("https://", "http://", "http:")):
                        raise ValueError(f"Invalid origin format: {origin_lower}")
                    processed.append(origin_lower)

            policy_parts.append(f"{feature}=({' '.join(processed)})")

        return ", ".join(policy_parts)



import time
import uuid
from contextvars import ContextVar

from fastapi import Request
from loguru import logger
from starlette.types import ASGIApp

from ash_utils.constants import REQUEST_ID_HEADER_NAME

request_id_var: ContextVar[str] = ContextVar("request_id_var", default="")


class RequestIDMiddleware:
    """
    Middleware responsible for contextualizing logger with request_id that
    helps to find all logs for a specific request.
    `request_id` is returned in the "X-Request-ID" header of the response.
    """

    def __init__(self, app: ASGIApp, header_name: str = REQUEST_ID_HEADER_NAME):
        self.app = app
        self.header_name = header_name

    async def __call__(self, scope, receive, send):
        if scope["type"] != "http":  # pragma: no cover
            await self.app(scope, receive, send)
            return

        request = Request(scope, receive)
        request_id = request.headers.get(self.header_name, str(uuid.uuid4()))
        request_id_var.set(request_id)

        with logger.contextualize(request_id=request_id):
            logger.info(f"Request started | Path: {request.url.path}")
            start_time = time.monotonic()

            async def send_wrapper(message):
                if message["type"] == "http.response.start":
                    headers = message.setdefault("headers", [])
                    headers.append((self.header_name.encode(), request_id.encode()))
                await send(message)

            try:
                await self.app(scope, receive, send_wrapper)
            finally:
                logger.info(
                    f"Request finished | Path: {request.url.path} | Duration: {time.monotonic() - start_time} s",
                )



from enum import StrEnum

from loguru import logger
from pydantic import BaseModel, Field

DEFAULT_LOG_MESSAGE = "New support ticket"


class LogLevel(StrEnum):
    TRACE = "TRACE"
    DEBUG = "DEBUG"
    INFO = "INFO"
    SUCCESS = "SUCCESS"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class PriorityLevel(StrEnum):
    """
    Constants that map to the priority levels in Zendesk
    """

    P0 = "p0"
    P1 = "p1"
    P2 = "p2"
    P3 = "p3"


class TicketType(StrEnum):
    """
    Constants that map to the ticket types in Zendesk
    """

    ESCALATE_ONCALL_EVENT_TO_OPS = "escalate-oncall-event-to-ops"
    ESCALATE_LAB_EVENT_KIT_ISSUE = "escalate-lab-event-kit-issue"


class SupportTicketDTO(BaseModel):
    kit_id: str
    ticket_type: TicketType
    subject: str
    message_body: str
    custom_fields: dict = Field(default_factory=dict)
    partner_id: str | None = None
    priority: PriorityLevel = PriorityLevel.P3


def create_support_ticket(
    message: str = DEFAULT_LOG_MESSAGE,
    *,
    ticket_data: SupportTicketDTO,
    log_level: LogLevel = LogLevel.ERROR,
):
    """
    This function logs a message along with a support ticket data using Loguru.
    The ticket data is attached as an extra field for better log searching and analysis.

    Args:
        message: Descriptive message about the support ticket event.
        ticket_data: SupportTicketDTO containing all relevant ticket information.
        log_level: Severity level for the log entry (defaults to ERROR).
                  Must be one of the values from the LogLevel enum.

    Example:
        >>> ticket = SupportTicketDTO(
        ...     kit_id="AW12345678",
        ...     ticket_type=TicketType.ESCALATE_LAB_EVENT_KIT_ISSUE,
        ...     partner_id="partner-123",
        ...     subject="Issue with kit processing",
        ...     message_body="Result is blocked by lab",
        ...     custom_fields={"lab_id": "123", "sample_type": "blood"}
        ... )
        >>> create_support_ticket("Some issue with the lab", ticket)
    """
    logger.log(log_level, message, support_ticket_data=ticket_data.model_dump(mode="json"))



from ash_utils.support.ticket import LogLevel, SupportTicketDTO, TicketType, create_support_ticket

__all__ = [
    "LogLevel",
    "SupportTicketDTO",
    "TicketType",
    "create_support_ticket",
]
